#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the GPLv3 license.

"""
Authenticates users using PAM.

Uses PAM (via the `simplepam <https://pypi.python.org/pypi/simplepam/>`_
module) to authenticate users using a PAM service. By default, the ``login``
service will be used, which typically authenticates user accounts for the
local machine. You may want to use the ``service`` option to change this:

.. code-block:: lighttpd
   :linenos:

    stores {
        pam.omni {
            service "omni"
        }
    }

With this snippet in the configuration file, the ``/etc/pam.d/omni``
`service definition file
<http://www.linux-pam.org/Linux-PAM-html/sag-configuration-file.html>`__
will be expected to be readable by OMNI.


**Configuration options:**

``service`` (optional)
    Name of the PAM service used to perform authentication. If not provided
    the ``login`` service is used by default.

``min-uid`` (optional)
    Users with an UID smaller that this value will not be useable. This is
    typically used to hide special system users from user listings. The
    default value is `1000`, which is a typical value for GNU/Linux systems
    and some BSD systems as well.
"""

from .. import store
from .. import valid


class PAMStore(store.Base):
    def __init__(self, service=None, min_uid=0):
        super(PAMStore, self).__init__()
        self._service = service
        self._min_uid = min_uid

    def authenticate(self, username, password):
        # XXX: Should we be picky and check for validity of the UID?
        import simplepam
        return simplepam.authenticate(username, password, self._service)

    def usernames(self):
        # FIXME: This works but it might be better to use NSS directly.
        import pwd
        return (pw.pw_name for pw in pwd.getpwall()
                if pw.pw_uid >= self._min_uid)


config_schema = valid.Schema({
    valid.Optional("service"): valid.Identifier,
    valid.Optional("min-uid"): valid.NaturalNumber,
})


@valid.argument(config_schema)
def from_config(config):
    return PAMStore(config.get("service", None), config.get("min-uid", 1000))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the MIT license.

import unittest
from .. import store


class TestAuthenticator(unittest.TestCase):

    def setUp(self):
        self.a = store.Authenticator()

    def test_authenticator_authenticate(self):
        with self.assertRaises(NotImplementedError):
            self.a.authenticate("bob", "b0b")

    def test_authenticator_usernames(self):
        with self.assertRaises(NotImplementedError):
            l = list(self.a.usernames())

    def test_authenticator_set_password(self):
        with self.assertRaises(NotImplementedError):
            self.a.set_password("bob", "b0b")

    def test_authenticator_has_user(self):
        with self.assertRaises(NotImplementedError):
            e = self.a.has_user("bob")

    def test_authenticator_get_user(self):
        with self.assertRaises(NotImplementedError):
            u = self.a.get_user("bob")

    def test_authenticator_create_user(self):
        with self.assertRaises(NotImplementedError):
            self.a.create_user("bob", "b0b")

    def test_authenticator_delete_user(self):
        with self.assertRaises(NotImplementedError):
            self.a.delete_user("bob")


class _RealmBaseTest(object):

    def test_realm_authenticate(self):
        with self.assertRaises(KeyError):
            self.r.authenticate("bob", "b0b")

    def test_realm_usernames(self):
        self.assertEquals([], list(self.r.usernames()))

    def test_realm_set_password(self):
        with self.assertRaises(KeyError):
            self.r.set_password("bob", "b0b")

    def test_realm_has_user(self):
        self.assertFalse(self.r.has_user("bob"))

    def test_realm_get_user(self):
        with self.assertRaises(KeyError):
            u = self.r.get_user("bob")

    def test_realm_create_user(self):
        with self.assertRaises(NotImplementedError):
            self.r.create_user("bob", "b0b")

    def test_realm_delete_user(self):
        with self.assertRaises(NotImplementedError):
            self.r.delete_user("bob")

    def test_realm_is_readonly(self):
        self.assertTrue(self.r.readonly)


class TestEmptyRealm(unittest.TestCase, _RealmBaseTest):
    def setUp(self):
        self.r = store.Realm("test realm")


class DummyStore(store.Base):
    def authenticate(self, username, password):
        raise KeyError(username)

    def usernames(self):
        return ()

    def set_password(self, username, password):
        raise KeyError(username)

    def has_user(self, username):
        return False

    def get_user(self, username):
        raise KeyError(username)


class WritableDummyStore(DummyStore):
    readonly = False


class TestRealmWithEmptyAuthenticator(unittest.TestCase, _RealmBaseTest):
    def setUp(self):
        self.r = store.Realm("test realm", DummyStore())


class TestRealmWithEmptyWritableAuthenticators(unittest.TestCase,
        _RealmBaseTest):
    def setUp(self):
        self.r = store.Realm("test realm", DummyStore(),
                WritableDummyStore())

    def test_realm_is_readonly(self):
        self.assertFalse(self.r.readonly)


class TestRealm(unittest.TestCase):
    def setUp(self):
        from ..stores import trivial
        self.r = store.Realm("trivial realm",
                trivial.TrivialStore("bob", "b0b"),
                trivial.TrivialStore("alice", "al1ce"))

    def test_authenticate_valid(self):
        self.assertTrue(self.r.authenticate("bob", "b0b"))
        self.assertTrue(self.r.authenticate("alice", "al1ce"))

    def test_authenticate_invalid(self):
        self.assertFalse(self.r.authenticate("bob", "b"))
        self.assertFalse(self.r.authenticate("alice", "a"))

    def test_authenticate_nonexistent(self):
        with self.assertRaises(KeyError):
            self.r.authenticate("kenny", "g")

    def test_usernames(self):
        self.assertEqual(["alice", "bob"], list(sorted(self.r.usernames())))

    def test_set_password(self):
        self.r.set_password("bob", "blurb")
        self.assertTrue(self.r.authenticate("bob", "blurb"))
        self.assertFalse(self.r.authenticate("bob", "b0b"))
        self.r.set_password("alice", "blarb")
        self.assertTrue(self.r.authenticate("alice", "blarb"))
        self.assertFalse(self.r.authenticate("alice", "al1ce"))

    def test_set_password_nonexistent(self):
        with self.assertRaises(KeyError):
            self.r.set_password("kenny", "g")

    def test_has_user(self):
        self.assertTrue(self.r.has_user("bob"))
        self.assertTrue(self.r.has_user("alice"))
        self.assertFalse(self.r.has_user("kenny"))

    def test_get_user(self):
        # FIXME: Change when stores implement getting user details.
        with self.assertRaises(NotImplementedError):
            u = self.r.get_user("bob")
        with self.assertRaises(NotImplementedError):
            u = self.r.get_user("alice")

    def test_get_user_nonexistent(self):
        with self.assertRaises(KeyError):
            u = self.r.get_user("kenny")

    def test_is_readonly(self):
        self.assertFalse(self.r.readonly)

    def test_create_user(self):
        with self.assertRaises(NotImplementedError):
            self.r.create_user("bob", "b0b")


class TestRealmDupUsers(TestRealm):
    def setUp(self):
        from ..stores import trivial
        self.r = store.Realm("trivial realm",
                trivial.TrivialStore("bob", "b0b"),
                trivial.TrivialStore("alice", "al1ce"),
                trivial.TrivialStore("alice", "al1ce2"))


class TestReadonlyStore(TestRealm):
    def setUp(self):
        super(TestReadonlyStore, self).setUp()
        for s in self.r:
            s.readonly = True

    def test_set_password(self):
        with self.assertRaises(store.AccessError):
            self.r.set_password("bob", "blurb")

    def test_is_readonly(self):
        self.assertTrue(self.r.readonly)


class TestOMNI(unittest.TestCase):
    def setUp(self):
        self.o = store.OMNI()

    def test_add_store(self):
        s = DummyStore()
        self.o.add_store("dummy.store", s)
        self.assertEqual(s, self.o.get_store("dummy.store"))

    def test_add_realm_and_store(self):
        s = DummyStore()
        self.o.add_store("dummy.store", s)
        r = store.Realm("default realm", s)

        self.o.add_realm("default", r)
        self.assertEqual(r, self.o.get_realm("default"))
        self.assertEqual(r, self.o["default"])
        self.assertTrue("default" in self.o)

        self.assertEqual(r, self.o.get_realm_or_store("default"))
        self.assertEqual(s, self.o.get_realm_or_store("dummy.store"))

    def test_add_store_duplicate(self):
        s = DummyStore()
        self.o.add_store("dummy.store", s)
        with self.assertRaises(KeyError):
            self.o.add_store("dummy.store", s)

    def test_add_realm_duplicate(self):
        s = DummyStore()
        self.o.add_store("dummy.store", s)
        r = store.Realm("default realm", s)
        self.o.add_realm("default", r)
        with self.assertRaises(KeyError):
            self.o.add_realm("default", r)

    def test_enumerate_stores(self):
        self.assertEqual([], list(self.o.stores))
        s = DummyStore()
        self.o.add_store("dummy.store", s)
        self.assertEqual([("dummy.store", s)], list(self.o.stores))

    def test_enumerate_realms(self):
        self.assertEqual([], list(self.o.realms))
        s = DummyStore()
        self.o.add_store("dummy.store", s)
        r = store.Realm("default realm", s)
        self.o.add_realm("default", r)
        self.assertEqual([("default", r)], list(self.o.realms))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the MIT license.

"""
Utilities for dealing with configuration files.
"""

from six import text_type, string_types
from schema import Schema, SchemaError
from schema import Optional, And, Or, Use
from functools import wraps
from os import path


_UNDEFINED = object()


def IPv4Address(value):
    from socket import inet_pton, AF_INET, error
    try:
        inet_pton(AF_INET, value)
        return True
    except error:
        return False


def IPv6Address(value):
    from socket import inet_pton, AF_INET6, error
    try:
        inet_pton(AF_INET6, value)
        return True
    except error:
        return False


def Match(regexp, flags=0):
    import re
    match = re.compile(regexp, flags).match
    @wraps(Match)
    def f(value):
        return match(value) is not None
    return f


Password       = Use(text_type)
Text           = And(text_type, len)
Identifier     = And(text_type, Match(r"^[_a-zA-Z]\w*$"),
                     error="Invalid identifier")
DotIdentifier  = And(text_type, Match(r"^[_\.a-zA-Z][\w\.]*$"),
                     error="Invalid identifier")
Path           = And(Or(text_type, *string_types), Use(path.realpath),
                     error="Invalid path")
Number         = Use(int)
NaturalNumber  = And(Number, lambda v: v >= 0, error="Not a positive number")
PortNumber     = And(Number, lambda v: 0 < v <= 65535)
Hostname       = Match(r"^[\w][\.\w]*$")
NetworkAddress = Or(IPv4Address, IPv6Address, Hostname)


def argument(schema, name=None):
    def decorate(func):
        if name is None:
            @wraps(func)
            def wrapper(data, *arg, **kw):
                return func(schema.validate(data, *arg, **kw))
        else:
            # TODO: Implement calling the function with validation done for the
            #       given named argument.
            raise NotImplementedError
        return wrapper
    return decorate



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the GPLv3 license.

from ..stores import plain
from .. import store
from textwrap import dedent
from six import StringIO
import unittest


class StringIOWithContext(StringIO, object):
    def __init__(self, persistent, data=None):
        super(StringIOWithContext, self).__init__(data)
        self.persistent = persistent

    def __enter__(self):
        return self

    def __exit__(self, type_, value, traceback):
        self.persistent.data = self.getvalue()


class PersistentStringIO(object):
    def __init__(self, data=""):
        self.data = data

    def __call__(self, mode):
        if "w" in mode:
            return StringIOWithContext(self)
        else:
            return StringIOWithContext(self, self.data)


class TestPlainStoreUnimplementeds(unittest.TestCase):

    def test_format_crypt_password(self):
        f = plain.BaseFileFormat(lambda mode: None)
        with self.assertRaises(NotImplementedError):
            f.crypt_password("user", "pass")


class TestPlainStoreFileFormat(unittest.TestCase):

    def setUp(self):
        self.last_io = None

    def tearDown(self):
        self.last_io = None

    def userdb_string(self, data):
        self.last_io = PersistentStringIO(dedent(data))
        return plain.PlainFileFormat(self.last_io)

    data01 = u"""\
    bob:b0b
    alice:4l1c3
    """

    def test_user_list(self):
        with self.userdb_string(self.data01) as db:
            self.assertEqual(list(db.users), ["bob", "alice"])

    def test_user_exists(self):
        with self.userdb_string(self.data01) as db:
            self.assertTrue("bob" in db)
            self.assertTrue(db.__contains__("bob"))
            self.assertFalse("peter" in db)
            self.assertFalse(db.__contains__("peter"))

    def test_user_get(self):
        with self.userdb_string(self.data01) as db:
            self.assertTrue(db["bob"])
            self.assertTrue(len(db["bob"]) > 0)
            self.assertEqual("b0b", db["bob"])

    def test_user_add(self):
        with self.userdb_string(self.data01) as db:
            db.add("andrew", "andr3w")
            self.assertTrue("andrew" in db)
            self.assertEqual("andr3w", db["andrew"])
        expected = dedent(self.data01).strip()
        expected += "\nandrew:andr3w\n"
        self.assertEqual(expected, self.last_io.data)

    def test_user_add_with_extrainfo(self):
        with self.userdb_string(self.data01) as db:
            db.add("andrew", "andr3w", "Some extra, ignored info")
            self.assertTrue("andrew" in db)
            self.assertEqual("andr3w", db["andrew"])
        expected = dedent(self.data01).strip()
        expected += "\nandrew:andr3w:Some extra, ignored info\n"
        self.assertEqual(expected, self.last_io.data)

    def test_user_add_existing(self):
        with self.userdb_string(self.data01) as db:
            with self.assertRaises(KeyError):
                db.add("bob", "meh")

    def test_user_delete(self):
        with self.userdb_string(self.data01) as db:
            db.delete("bob")
            self.assertFalse("bob" in db)
        self.assertEqual("alice:4l1c3\n", self.last_io.data)

    def test_user_delete_nonexisting(self):
        with self.userdb_string(self.data01) as db:
            with self.assertRaises(KeyError):
                db.delete("peter")
        expected = dedent(self.data01)
        self.assertEqual(expected, self.last_io.data)

    def test_user_crypt_password(self):
        with self.userdb_string(self.data01) as db:
            self.assertEqual("b0b", db.crypt_password("bob", "b0b", db["bob"]))

    data02 = u"""\
    peter:p3t3r:Extra info
    bob:b0b:
    alice:4l1c3
    """

    def test_extrainfo_roundtrip(self):
        with self.userdb_string(self.data02) as db:
            pass
        self.assertEqual(dedent(self.data02), self.last_io.data)

    def test_list_usernames(self):
        with self.userdb_string(self.data02) as db:
            self.assertEqual(["peter", "bob", "alice"], list(db.users))


class TestHtpasswdStoreFormat(unittest.TestCase):

    def setUp(self):
        self.last_io = None

    def tearDown(self):
        self.last_io = None

    def userdb_string(self, data, method):
        m = plain._crypt_methods.get(method, None)
        if m is None:
            self.skipTest(("Crypt method '{}' is not supported by this"
                " Python version").format(method))
        self.last_io = PersistentStringIO(dedent(data))
        return plain.HtpasswdFileFormat(self.last_io, m)


    # Generated with: printf "alice:$(openssl passwd -crypt mirror)" 
    data01 = (u"""\
    alice:su7aWQyEG4lo.
    """, "crypt")

    def test_user_crypt_check_password(self):
        with self.userdb_string(*self.data01) as db:
            self.assertEqual(db["alice"],
                    db.crypt_password("alice", "mirror", db["alice"]))

    def test_user_crypt_check_password_invalid(self):
        with self.userdb_string(*self.data01) as db:
            self.assertNotEqual(db["alice"],
                    db.crypt_password("alice", "hatter", db["alice"]))

    def test_user_crypt_newpassword(self):
        with self.userdb_string(*self.data01) as db:
            old_crypted_pw = db["alice"]
            new_crypted_pw = db.crypt_password("alice", "mirror")
            # If we don't pass the salt (or the existing crypted password),
            # a new crupted password one will be created with a new salt.
            self.assertNotEqual(old_crypted_pw, new_crypted_pw)

    def test_list_usernames(self):
        with self.userdb_string(*self.data01) as db:
            self.assertEqual(["alice"], list(db.users))


class TestPlainStoreInstatiation(unittest.TestCase):

    valid_confs = (
        { "path": "path/to/passwd" },
        { "path": "path/to/passwd", "format": "plain" },
        { "path": "path/to/passwd", "format": "htpasswd" },
        { "path": "path/to/passwd", "format": "htpasswd", "method": "crypt" },
    )

    def test_from_valid_conf(self):
        for conf in self.valid_confs:
            s = plain.from_config(conf)
            self.assertTrue(isinstance(s, store.Base))

    invalid_confs = (
        # No "path" key.
        { },
        # Invalid "format" values.
        { "path": "path/to/passwd", "format": "" },
        { "path": "path/to/passwd", "format": 42 },
        { "path": "path/to/passwd", "format": True },
        { "path": "path/to/passwd", "format": 3.14 },
        { "path": "path/to/passwd", "format": "bogus" },
        # Invalid "crypt" methods for htpasswd format
        { "path": "path/to/passwd", "format": "htpasswd", "method": "" },
        { "path": "path/to/passwd", "format": "htpasswd", "method": 42 },
        { "path": "path/to/passwd", "format": "htpasswd", "method": True },
        { "path": "path/to/passwd", "format": "htpasswd", "method": 3.14 },
        { "path": "path/to/passwd", "format": "htpasswd", "method": "bogus" },
        # The configuration dictionary is strictly checked, and keys
        # which are of not interest for instantiation cause errors.
        { "path": "path/to/passwd", "pi": 3.14 },
    )

    def test_from_invalid_conf(self):
        for conf in self.invalid_confs:
            with self.assertRaises(Exception):
                s = plain.from_config(conf)
                self.assertTrue(isinstance(s, plain.PlainStore))

    def test_find_store(self):
        self.assertEqual(plain, store.find("plain"))


class TestablePlainStore(plain.PlainStore):
    def __init__(self, data, path, format_, *fargs):
        super(TestablePlainStore, self).__init__(path, format_, *fargs)
        self.last_io = PersistentStringIO(dedent(data))

    def _open_file(self, mode):
        return self.last_io(mode)


class TestPlainStoreAuthentication(unittest.TestCase):

    plain_data = TestPlainStoreFileFormat.data01
    crypt_data = TestHtpasswdStoreFormat.data01[0]

    def test_plain_authenticate_success(self):
        s = TestablePlainStore(self.plain_data, "path/to/file",
                plain.PlainFileFormat)
        self.assertTrue(s.authenticate("bob", "b0b"))

    def test_plain_authenticate_failure(self):
        s = TestablePlainStore(self.plain_data, "path/to/file",
                plain.PlainFileFormat)
        self.assertFalse(s.authenticate("bob", "badpass"))

    def test_plain_list_usernames(self):
        s = TestablePlainStore(self.plain_data, "path/to/file",
                plain.PlainFileFormat)
        self.assertEqual(["alice", "bob"], list(sorted(s.usernames())))

    def test_plain_set_password(self):
        s = TestablePlainStore(self.plain_data, "path/to/files",
                plain.PlainFileFormat)
        self.assertTrue(s.authenticate("bob", "b0b"))
        s.set_password("bob", "blurb")
        self.assertFalse(s.authenticate("bob", "b0b"))
        self.assertTrue(s.authenticate("bob", "blurb"))

    def test_plain_set_password_invalid_user(self):
        s = TestablePlainStore(self.plain_data, "path/to/files",
                plain.PlainFileFormat)
        self.assertFalse(s.has_user("peter"))
        with self.assertRaises(KeyError):
            s.set_password("peter", "parker")

    def test_crypt_set_password(self):
        m = plain._crypt_methods.get("crypt", None)
        if m is None:
            self.skipTest("Crypt method 'crypt' is not supported by this"
                " Python version")
            return

        s = TestablePlainStore(self.crypt_data, "path/to/file",
                plain.HtpasswdFileFormat, m)
        s.set_password("alice", "shiny")
        self.assertFalse(s.authenticate("alice", "mirror"))
        self.assertTrue(s.authenticate("alice", "shiny"))

    def test_create_user(self):
        s = TestablePlainStore(self.plain_data, "path/to/file",
                plain.PlainFileFormat)
        s.create_user("peter", "p3t3r")
        self.assertTrue(s.has_user("peter"))
        self.assertTrue(s.authenticate("peter", "p3t3r"))

    def test_delete_user(self):
        s = TestablePlainStore(self.plain_data, "path/to/file",
                plain.PlainFileFormat)
        s.delete_user("bob")
        self.assertFalse(s.has_user("bob"))

    def test_delete_nonexisting_user(self):
        s = TestablePlainStore(self.plain_data, "path/to/file",
                plain.PlainFileFormat)
        with self.assertRaises(KeyError):
            s.delete_user("peter")

    def test_create_existing_user(self):
        s = TestablePlainStore(self.plain_data, "path/to/file",
                plain.PlainFileFormat)
        with self.assertRaises(KeyError):
            s.create_user("bob", "b0b")

    def test_create_user_with_extrainfo(self):
        s = TestablePlainStore("", "path/to/file",
                plain.PlainFileFormat)
        s.create_user("bob", "b0b", role="manager")
        self.assertTrue(s.last_io.data.startswith("bob:b0b:"))


NOT_A_FILE = "/tmp/this/is/a/very/long/path/to/a/nonexistent/file"

class TestPlainStoreOpen(unittest.TestCase):
    def test_open_nonexistent_file(self):
        s = plain.from_config({ "path": NOT_A_FILE })
        self.assertEqual([], list(s.usernames()))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the GPLv3 license.

"""
Authenticate users using plain text files.

Implements a store that loads user and passwords from plain text files.
The recognized file format is line-based and compatible with a number of
other text-based user databases. The parser is intentionally forgiving with
unrecognized input lines.

One of the goals of the store is to be interoperable with plain-text user
files used by third-party software, so their files can be shared with OMNI.
For example, the following configures a store which can access ``htpasswd``
files, as typically used by web servers:

.. code-block:: lighttpd
   :linenos:

    stores {
        plain.htpasswd {
            format "htpasswd"
            path "/etc/httpd/htpasswd"
        }
    }


**Configuration options:**

``format`` (optional)

    Format of the text file. Valid values are:

    - ``plain``: This is the default value. Passwords are stored in plain
      form (not hashed).

      **Warning**: please assign permissions accordingly in the file system
      to avoid users to eavesdrop!

    - ``unix``: The passwords are hashed, using the configured ``method``
      to hash the passwords.

    - ``htpasswd``: This is an alias for ``unix`` that configures ``method``
      to be ``crypt``.

``method`` (optional)

    Hashing method using to store the passwords. The following values are
    valid:

    - ``crypt``: Use the standard Unix ``crypt(3)`` hashing, basing on the
      DES algorithm.

      **Warning**: Note that this is method is old and should be avoided
      if possible; it is mainly provided for compatibility with other tools
      and because with Python 2.x is the only available method.

    - ``md5``: Use MD5 for password hashing.

    - ``sha256``: Use SHA256 for password hasshing.

    - ``sha512``: Use SHA512 for password hashing. This is the recommended
      setting.
"""

from .. import store, valid
from six import iteritems, iterkeys, StringIO
from errno import ENOENT
from codecs import open
from collections import OrderedDict
from base64 import b64encode, b64decode
from six.moves.cPickle import loads as unpickle, dumps as pickle
import crypt, sys

_crypt_methods = {}
if hasattr(crypt, "methods"):  # pragma: no cover
    for _method in crypt.methods:
        _crypt_methods[_method.name.lower()] = _method
    _crypt_mksalt = crypt.mksalt
else:  # pragma: no cover
    # Patch up missing parts from the crypt module from Python 2.x
    _crypt_methods["crypt"] = "crypt"
    class _crypt_mksalt(object):
        from string import digits, ascii_letters
        _salt_chars = digits + ascii_letters

        def __call__(self, method=None):
            from random import choice
            return choice(self._salt_chars) + choice(self._salt_chars)
    _crypt_mksalt = _crypt_mksalt()

if sys.version_info.major == 2:  # pragma: no cover
    from hmac import compare_digest as _compare_digest_base
    def compare_digest(a, b):
        if isinstance(a, unicode):
            a = a.encode("utf-8")
        if isinstance(b, unicode):
            b = b.encode("utf-8")
        return _compare_digest_base(a, b)
else:  # pragma: no cover
    from hmac import compare_digest


def serialize_info(value):
    return "p" + str(b64encode(pickle(value)))


def unserialize_info(data):
    if data[0] == "p":
        return unpickle(b64decode(data[1:]))
    return data


class BaseFileFormat(object):
    def __init__(self, open_func):
        self._users = OrderedDict()
        self._infos = {}
        self._open = open_func

    def __contains__(self, username):
        return username in self._users

    def __getitem__(self, username):
        return self._users[username]

    def __setitem__(self, username, password):
        self._users[username] = password

    @property
    def users(self):
        return iterkeys(self._users)

    def add(self, username, password, extrainfo=None):
        if username in self._users:
            raise KeyError("User already exists: {}".format(username))
        self._users[username] = self.crypt_password(username, password)
        if extrainfo:
            assert username not in self._infos
            self._infos[username] = extrainfo
        return self

    def delete(self, username):
        if username in self._users:
            del self._users[username]
            if username in self._infos:
                del self._infos[username]
        else:
            raise KeyError("User not found: {}".format(username))

    def __enter__(self):
        with self._open("r") as fd:
            for line in fd.readlines():
                line = line.strip().split(":", 2)
                self._users[line[0]] = line[1]
                if len(line) == 3:
                    self._infos[line[0]] = line[2]
            return self

    def __exit__(self, type_, value, traceback):
        with self._open("w") as fd:
            for u, p in iteritems(self._users):
                x = self._infos.get(u)
                if x is None:
                    fd.write("{}:{}\n".format(u, p))
                else:
                    fd.write("{}:{}:{}\n".format(u, p, x))
        return False

    def crypt_password(self, username, password, salt=None):
        raise NotImplementedError


class PlainFileFormat(BaseFileFormat):
    def crypt_password(self, username, password, salt=None):
        return password


class HtpasswdFileFormat(BaseFileFormat):
    def __init__(self, open_func, method):
        super(HtpasswdFileFormat, self).__init__(open_func)
        self._method = method

    def crypt_password(self, username, password, salt=None):
        if salt is None:
            salt = _crypt_mksalt(self._method)
        return crypt.crypt(password, salt)


class _ContextStringIO(StringIO, object):
    def __enter__(self):
        return self
    def __exit__(self, type_, value, traceback):
        pass


class PlainStore(store.Base):
    readonly = False

    def __init__(self, path, format_, *fargs):
        super(PlainStore, self).__init__()
        self._format = format_
        self._fargs = fargs
        self._path = path

    def _open_file(self, mode):  # pragma: no cover
        try:
            return open(self._path, mode, encoding="utf-8")
        except IOError as e:
            if e.errno == ENOENT:
                return _ContextStringIO("")
            else:
                raise e

    def authenticate(self, username, password):
        with self._format(self._open_file, *self._fargs) as db:
            current = db[username]
            crypted = db.crypt_password(username, password, current)
            return compare_digest(current, crypted)

    def usernames(self):
        with self._format(self._open_file, *self._fargs) as db:
            return db.users

    def set_password(self, username, password):
        with self._format(self._open_file, *self._fargs) as db:
            if username not in db:
                raise KeyError("invalid username {}".format(username))
            db[username] = db.crypt_password(username, password)

    def create_user(self, username, password, **kw):
        with self._format(self._open_file, *self._fargs) as db:
            db.add(username, password, serialize_info(kw) if kw else None)

    def delete_user(self, username):
        with self._format(self._open_file, *self._fargs) as db:
            db.delete(username)


config_schema = valid.Schema({
    "path": valid.Path,
    valid.Optional("format"): valid.Or(u"plain", u"unix", u"htpasswd"),
    valid.Optional("method"): valid.Or(*tuple(_crypt_methods.keys())),
})


@valid.argument(config_schema)
def from_config(config):
    file_format = config.get("format", "plain")
    if file_format == "plain":
        file_format = PlainFileFormat
        format_args = ()
    elif file_format in ("htpasswd", "unix"):
        file_format = HtpasswdFileFormat
        crypt_method = _crypt_methods[config.get("method", "crypt")]
        format_args = (crypt_method,)
    else:
        raise KeyError("File format not supported: {}".format(file_format))

    return PlainStore(config["path"], file_format, *format_args)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the GPLv3 license.

"""
Credential store definition and related utilities.
"""
from inspect import isgenerator
from six import iteritems


class Authenticator(object):
    readonly = True

    """
    Defines the basic interface for authenticating users.
    """
    def authenticate(self, username, password):
        """
        Authenticates a given user

        :username: Name of the user.
        :password: Password for the user.
        """
        raise NotImplementedError

    def usernames(self):
        """
        Returns an iterator yielding known user names.
        """
        raise NotImplementedError

    def set_password(self, username, password):
        """
        Set a new password for the given user.
        """
        raise NotImplementedError

    def has_user(self, username):
        """
        Checks whether the store knows about a given user.
        """
        try:
            return self.get_user(username) is not None
        except NotImplementedError:
            for u in self.usernames():
                if u == username:
                    return True
            return False

    def get_user(self, username):
        """
        Obtains the details for a given user name.
        """
        raise NotImplementedError

    def create_user(self, username, password, **kw):
        """
        Creates a new user name.

        :param username: Name of the user.
        """
        raise NotImplementedError

    def delete_user(self, username):
        """
        Deletes an user.

        :param username: Name of the user.
        """
        raise NotImplementedError


class AccessError(Exception):
    """
    Raised when a store or realm cannot be accessed in the requested way.
    """
    pass


class Realm(Authenticator, list):
    """
    A realm is a list of `Authenticator` objects which are tried in order.
    """
    def __init__(self, description, *realms):
        if len(realms) == 1 and isgenerator(realms[0]):
            super(Realm, self).__init__()
            self.extend(realms[0])
        else:
            super(Realm, self).__init__(realms)
        self.description = description

    @property
    def readonly(self):
        for a in self:
            if not a.readonly:
                return False
        return True

    def authenticate(self, username, password):
        for a in self:
            if a.has_user(username):
                return a.authenticate(username, password)
        raise KeyError("user {} does not exist".format(username))

    def usernames(self):
        seen = set()
        for a in self:
            for username in a.usernames():
                if username in seen:
                    continue
                seen.add(username)
                yield username

    def set_password(self, username, password):
        for a in self:
            if a.has_user(username):
                if a.readonly:
                    raise AccessError("user {} belongs is in a read-only store"
                            .format(username))
                a.set_password(username, password)
                return
        raise KeyError("user {} does not exist".format(username))

    def has_user(self, username):
        for a in self:
            if a.has_user(username):
                return True
        return False

    def get_user(self, username):
        for a in self:
            if a.has_user(username):
                return a.get_user(username)
        raise KeyError("user {} does not exist".format(username))

    def create_user(self, username, password, **kw):
        for a in self:
            if a.readonly:
                continue
            try:
                return a.create_user(username, password, **kw)
            except NotImplementedError:
                pass
        raise NotImplementedError

    def delete_user(self, username):
        for a in self:
            if a.readonly:
                continue
            try:
                return a.delete_user(username)
            except NotImplementedError:
                pass
        raise NotImplementedError


class Base(Authenticator):
    """
    Base class for authentication stores.
    """


class OMNI(object):
    def __init__(self):
        self._stores = {}
        self._realms = {}

    def add_store(self, name, store):
        if name in self._stores:
            raise KeyError("Store {} is already present".format(name))
        assert isinstance(store, Base)
        self._stores[name] = store

    def add_realm(self, name, realm):
        if name in self._realms:
            raise KeyError("Realm {} is already present".format(name))
        assert isinstance(realm, Realm)
        self._realms[name] = realm

    def get_realm(self, name):
        return self._realms[name]

    def get_store(self, name):
        return self._stores[name]

    def get_realm_or_store(self, name):
        if "." in name:
            return self.get_store(name)
        else:
            return self.get_realm(name)

    def __contains__(self, name):
        return name in self._realms

    def __getitem__(self, name):
        return self._realms[name]

    @property
    def stores(self):
        """
        Provides an iterator over `(name, store)` pairs.
        """
        return iteritems(self._stores)

    @property
    def realms(self):
        """
        Provides an iterator over `(name, realm)` pairs.
        """
        return iteritems(self._realms)



def find(storename):
    from importlib import import_module
    import omni.stores
    return import_module("." + storename, "omni.stores")



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the GPLv3 license.

"""
Usage: omni [--version] [--config=<path>] <command> [<args>...]

Options:

  -C, --config=PATH  Path to comfiguration file [default: /etc/omni.conf].
  --version          Show the OMNI version and exit.
  -h, --help         Show this help message.

The most commonly used commands are:

  server           Starts the OMNI server.
  list-stores      Lists the configured stores.
  list-realms      Lists the configured realms.
  list-users       Lists users in realm or a store.
  create-user      Adds an user to a realm or a store.
  delete-user      Deletes an user from a realm or a store.
  change-password  Changes the passform for an user.
  help             Show help about different topics.

See 'omni help <command>' for more information on a specific command.
"""

from .metadata import metadata
from . import app, valid
from six import iteritems, print_
from textwrap import dedent
import wcfg


def cmd_list_stores(omni_app):
    """
    Usage: omni list-stores

    Shows a list of configured stores.

    Options:

      -h, --help  Show this help message.
    """
    return sorted((name for name, _ in omni_app.stores))


def cmd_list_realms(omni_app):
    """
    Usage: omni list-realms

    Shows a list of configured realms.

    Options:

      -h, --help  Show this help message.
    """
    return sorted((name for name, _ in omni_app.realms))


def cmd_list_users(omni_app, realm_or_store):
    """
    Usage: omni list-users <realm-or-store>

    Shows a list of users in a realm or store.

    Options:

      -h, --help  Show this help message.
    """
    try:
        return sorted(omni_app.get_realm_or_store(realm_or_store).usernames())
    except KeyError:
        return "{}: invalid realm/store".format(realm_or_store)


def cmd_create_user(omni_app, realm_or_store, username, opt_stdin=False,
        opt_prompt_password=False):
    """
    Usage: omni create-user [options] <realm-or-store> <username>

    Adds a new user to a realm or a store.

    Options:

      -p, --prompt-password  Prompt for a password.
      -i, --stdin            Read password from standard input.
      -h, --help             Show this help message.
    """
    try:
        db = omni_app.get_realm_or_store(realm_or_store)
    except KeyError:
        return "{}: invalid realm/store".format(realm_or_store)

    if opt_stdin:
        from sys import stdin
        new_pw = stdin.readline().strip()
    elif opt_prompt_password:
        from getpass import getpass
        old_pw = getpass("password for {}: ".format(username))
        new_pw = getpass("password for {} (again): ".format(username))
        if old_pw != new_pw:
            return "passwords do not match"
    else:
        # TODO: Generate passwords and report them somehow. Alternatively,
        #       produce an activation link when we have that plug-in.
        new_pw = "#*" + username + "*#"

    try:
        return db.create_user(username, password=new_pw)
    except NotImplementedError:
        return "{}: user creation not suported".format(realm_or_store)


def cmd_delete_user(omni_app, realm_or_store, username):
    """
    Usage: omni delete-user <realm-or-store> <username>

    Deletes an user from a realm or a store.

    Options:

      -h, --help  Show this help message.
    """
    try:
        db = omni_app.get_realm_or_store(realm_or_store)
    except KeyError:
        return "{}: invalid realm/store".format(realm_or_store)
    try:
        db.delete_user(username)
    except KeyError:
        return "{}: unknown user".format(realm_or_store)
    except NotImplementedError:
        return "{}: user deletion not supported".format(realm_or_store)

def cmd_try_authenticate(omni_app, realm_or_store, username):
    """
    Usage: omni try-authenticate <realm-or-store> <username>

    Tries to authenticate an user interactively agains a given realm
    or a particular store. A zero exit status means that authentication
    succeeded. This command can be useful to debug configurations.

    Options:

      -h, --help  Show this help message.
    """
    from getpass import getpass
    try:
        authenticate = omni_app.get_realm_or_store(realm_or_store).authenticate
    except KeyError:
        return "{}: invalid realm/store".format(realm_or_store)

    if authenticate(username, getpass("password for {}: ".format(username))):
        return 0
    return 1


def cmd_change_password(omni_app, realm_or_store, username, opt_stdin=False):
    """
    Usage: omni change-password <realm-or-store> <username>

    Options:

      -i, --stdin  Read passwords from standard input. The first line must
                   be the existing password, and the second line the new
                   password.
      -h, --help   Show this help message.
    """
    from getpass import getpass
    try:
        db = omni_app.get_realm_or_store(realm_or_store)
    except KeyError:
        return "{}: invalid realm/store".format(realm_or_store)

    if opt_stdin:
        from sys import stdin
        old_pw = stdin.readline().strip()
        if not db.authenticate(username, old_pw):
            return "Incorrect password"
        new_pw = stdin.readline().strip()
    else:
        old_pw = getpass("old password for {}: ".format(username))
        if not db.authenticate(username, old_pw):
            return "Incorrect password"
        old_pw = getpass("new password for {}: ".format(username))
        new_pw = getpass("new password for {} (again): ".format(username))
        if old_pw != new_pw:
            return "Passwords do not match"
    db.set_password(username, new_pw)


config_http_schema = valid.Schema({
    valid.Optional("host"): valid.NetworkAddress,
    valid.Optional("port"): valid.PortNumber,
})


def cmd_server(omni_config, http_port=None, http_host=None, debugger=False):
    """
    Usage: omni server [--http-port=PORT] [--http-host=HOST] [--debugger]

    Options:

      -p, --http-port=PORT  Port in which to serve HTTP requests.
      --http-host=HOST      Host name of IP address to bind to.
      --debugger            Enable the debugger [default: False].

      -h, --help            Show this help message.
    """
    from aiowsgi.compat import asyncio
    import aiowsgi
    import logging
    logging.basicConfig()

    # Apply command line overrides.
    if http_host is None or http_port is None:
        http_config = { "port": 8080, "host": "localhost" }
        if "http" in omni_config:
            conf = config_http_schema.validate(omni_config["http"])
            http_config.update(conf)
        if http_host is None:
            http_host = http_config["host"]
        if http_port is None:
            http_port = http_config["port"]

    omni = app.make_application(omni_config)
    wsgi_app = app.make_wsgi_application(omni)

    if debugger:
        from backlash import DebuggedApplication
        from webob import Request
        from wsgiref.simple_server import make_server

        def getctx(e=None):
            req = None if e is None else Request(e)
            return { "env": e, "omni": omni, "req": req, "app": wsgi_app }

        wsgi_app = DebuggedApplication(wsgi_app,
                context_injectors=[getctx],
                console_init_func=getctx)
        make_server(http_host, http_port, wsgi_app).serve_forever()
    else:
        loop = asyncio.get_event_loop()
        aiowsgi.create_server(wsgi_app, loop=loop,
                host=http_host, port=http_port)
        try:
            loop.run_forever()
        except KeyboardInterrupt:
            pass


def cmd_help(commands, topics, topic):
    """
    Usage: omni help (topics | commands | <topic>)

    Shows information about a commands or about a help topic.

    Options:

      -h, --help  Show this help message.

    """
    if topics:
        return 0
    if commands:
        return sorted((k[4:].replace("_", "-")
            for k, v in iteritems(globals())
            if k != "cmd_help" and k.startswith("cmd_") and callable(v)))

    command = globals().get("cmd_" + topic.replace("-", "_"), None)
    if command is None:
        return "topic or command does not exist: {}".format(topic)

    from docopt import docopt as do_docopt
    do_docopt(dedent(command.__doc__), argv=[topic, "--help"])


class ArgBag(dict):
    def __getattr__(self, name):
        return self[name]

def docopt(doc, *args, **kwarg):
    from docopt import docopt as do_docopt

    args = do_docopt(dedent(doc), *args, **kwarg)
    result = ArgBag()
    for k, v in iteritems(args):
        if k.startswith("--"):
            k = "opt_" + k[2:]
        elif k.startswith("<") and k.endswith(">"):
            k = k[1:-1]
        result[k.lower().replace("-", "_")] = v
    return result


def iterfargs(func):
    from inspect import getargspec
    from six.moves import zip
    args, varargs, varkw, defaults = getargspec(func)
    num_defaults = 0 if defaults is None else len(defaults)

    if num_defaults > 0:
        for var in args[:-num_defaults]:
            yield (var, None, False, False)
        for var, default in zip(args[-num_defaults:], defaults):
            yield (var, default, False, False)
    else:
        for var in args:
            yield (var, None, False, False)
    if varargs is not None:
        yield (varargs, (), True, False)
    if varkw is not None:
        yield (varkw, {}, False, True)


def main():
    args = docopt(__doc__, version=metadata.version, options_first=True)
    command = globals().get("cmd_" + args.command.replace("-", "_"), None)
    if command is None:
        raise SystemExit("'{}' is not a valid command, see 'omni help'."
                .format(args.command))

    # Parse command line options for the subcommand.
    cmd_args = docopt(command.__doc__, argv=[args.command] + args.args)
    try:
        omni_config = None
        call_args = {}
        for name, default, is_args, is_kwarg in iterfargs(command):
            alt_name = "opt_" + name
            if is_args or is_kwarg:
                pass
            elif name == "omni_config":
                if omni_config is None:
                    omni_config = app.load_config(args.opt_config)
                call_args[name] = omni_config
            elif name == "omni_app":
                if omni_config is None:
                    omni_config = app.load_config(args.opt_config)
                call_args[name] = app.make_application(omni_config)
            elif alt_name in cmd_args:
                call_args[name] = cmd_args[alt_name]
            elif name in cmd_args:
                call_args[name] = cmd_args[name]
            else:
                call_args[name] = default

        result = command(**call_args)
    except wcfg.ParseError as e:
        raise SystemExit("error parsing config file:\n{s}:{!s}"
                .format(args.opt_config, e))
    except valid.SchemaError as e:
        raise SystemExit("configuration file validation error:\n{!s}"
                .format(e))

    if result is not None:
        from inspect import isgenerator
        if isgenerator(result) or isinstance(result, (list, tuple)):
            for item in result:
                print_(item)
        else:
            raise SystemExit(result)


if __name__ == "__main__":
    main()




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the GPLv3 license.

"""
Exposes a REST-ish authentication interface to OMNI.
"""

from webob.exc import HTTPNotFound
from webob.static import FileApp
from .web import routing
from . import valid, store
from Stache import Stache
from os import path, listdir

ASSETS_DIR = path.abspath(path.join(path.dirname(__file__), \
        "web", "assets"))
TEMPLATE_DIR = path.abspath(path.join(path.dirname(__file__), \
        "web", "templates"))


class StacheCache(Stache):
    def __init__(self):
        super(StacheCache, self).__init__()
        for name in listdir(TEMPLATE_DIR):
            if name.endswith(".mustache"):
                with open(path.join(TEMPLATE_DIR, name), "rU") as f:
                    name = name[:-len(".mustache")]
                    self.add_template(name, f.read())

    def __call__(self, data, template):
        return self.render_template(template, data)

render_stache = StacheCache()


change_pass_schema = valid.Schema({
    "username": valid.And(valid.Identifier,
        error="Invalid user name"),
    "password-old": valid.And(valid.Password, len,
        error="The password cannot be empty"),
    "password-new": valid.And(valid.Password, len,
        error="The password cannot be empty"),
    "password-new-again": valid.And(valid.Password, len,
        error="The password cannot be empty"),
})


class ChangePassword(routing.Routes):

    @routing.route("/{realm:dotid}/change-password", ("GET", "POST"))
    @routing.after(render_stache, "changepassword")
    def change_password_form(self, request, realm):
        data = {
            "pagetitle" : u"Password change",
            "form_url"  : self.url(realm=realm),
            "realm"     : realm,
        }
        realm = self.any_realm(realm)
        if request.method == "POST":
            data["error"] = True
            try:
                params = dict(request.POST.items())
                data["username"] = params["username"]
                params = change_pass_schema.validate(params)
                if not realm.authenticate(params["username"],
                        params["password-old"]):
                    data["message"] = u"Invalid username or password"
                elif not len(params["password-new"]):
                    data["message"] = u"New password cannot be empty"
                elif params["password-new"] != params["password-new-again"]:
                    data["message"] = u"Entered passwords do not match"
                elif params["password-old"] == params["password-new"]:
                    data["message"] = u"New password is the same as the old one"
                else:
                    realm.set_password(params["username"],
                            params["password-new"])
                    data["message"] = u"Password update successfully"
                    data["error"] = False
            except KeyError as e:
                data["message"] = e.message
            except valid.SchemaError as e:
                data["message"] = e.message
            except store.AccessError as e:
                data["message"] = e.message

        return data



class WSGIApplication(routing.Dispatcher):
    def __init__(self, omni):
        super(WSGIApplication, self).__init__()
        self._omni = omni
        self.add_route(self.do_auth)
        self.add_route(self.assets)
        self.plug_routes(ChangePassword().routes)

    def any_realm(self, realm):
        try:
            return self._omni.get_realm_or_store(realm)
        except KeyError:
            raise HTTPNotFound()

    def admin_realm(self, realm):
        return self._omni["omni-admin"]

    @routing.get("/+assets/{filename:str}")
    def assets(self, request, filename):
        return FileApp(path.join(ASSETS_DIR, filename))

    @routing.get("/{realm:dotid}/authenticate")
    @routing.authenticate(any_realm)
    def do_auth(self, request):
        request.response.content_type = "text/plain"
        return "OK\r\n"

    __call__ = routing.Dispatcher.dispatch_wsgi



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the GPLv3 license.

from codecs import open
from textwrap import dedent
from email import parser, message
from os import path
from re import match


class Metadata(message.Message, object):
    @staticmethod
    def get_author_field(name, field):
        return match(r"^([^<]+)\s+<([^>]*)>$", name).group(field)
    @staticmethod
    def get_multiline(content):
        return [line.strip() for line in content.splitlines() if line]
    @staticmethod
    def get_sub_keyed(content):
        m = parser.Parser().parsestr(content.strip())
        return dict((k, Metadata.get_multiline(m[k])) for k in m.keys())

    description = property(lambda self: self["Description"])
    version = property(lambda self: self["Version"])
    package = property(lambda self: self["Package"])
    url = property(lambda self: self["URL"])
    license = property(lambda self: self["License"])
    main_author_name = property(lambda self:
            self.get_author_field(self.authors[0], 1))
    main_author_email = property(lambda self:
            self.get_author_field(self.authors[0], 2))
    test_requirements = property(lambda self:
            self.get_multiline(self["Test-Requirements"]))
    extra_requirements = property(lambda self:
            self.get_sub_keyed(dedent(self["Extra-Requirements"])))
    entry_points = property(lambda self:
            self.get_multiline(self["Scripts"]))
    requirements = property(lambda self:
            self.get_multiline(self["Requirements"]))
    classifiers = property(lambda self:
            self.get_multiline(self["Classifiers"]))
    authors = property(lambda self: self.get_multiline(self["Authors"]))


def metadata():
    with open(path.join(path.dirname(__file__), "META"), "rU",
            encoding="utf-8") as f:
        return parser.Parser(Metadata).parse(f)
metadata = metadata()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the MIT license.

"""
Logging utilities.
"""

class WSGIRequestLogMiddleware(object):
    """
    Simple WSGI middleware to perform request logging.

    Logs each WSGI
    """
    def __init__(self, application):
        self._application = application

    def start_response(self, status, header, exc_info=None):
        pass

    def __call__(self, environ, start_response):
        return self._application(environ, start_response)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the MIT license.

from ..web import routing
from .. import valid
from webob import Request
from webob.exc import HTTPNotFound, HTTPMethodNotAllowed
from six import iteritems
import unittest
import re

sanitize_ident_re_sub = re.compile(r"[^_a-zA-Z0-9]").sub
sanitize_ident = lambda v: \
        str(sanitize_ident_re_sub("_", v).encode("ascii", "replace"))


class TestCamelCaseConverter(unittest.TestCase):

    def test_camel_case_to_unserscored(self):
        data = (
            (u"CamelCase", u"camel_case"),
            (u"camelCaseInTheMiddle", u"camel_case_in_the_middle"),
            (u"nocasechange", "nocasechange"),
            (u"keep_under_scores", "keep_under_scores"),
            (u"MultipleCAPSTogether", "multiple_caps_together"),
            (u"CAPSTogetherPrefix", u"caps_together_prefix"),
            (u"SuffixTogetherCAPS", u"suffix_together_caps"),
        )
        for camel_case, under_score in data:
            self.assertEqual(under_score,
                    routing.camel_to_under(camel_case))


class TestRoute(unittest.TestCase):

    def setUp(self):
        self.r = routing.Route(self.callback,
                u"static/{variable}/morestatic",
                "GET", "testroute")
        self.callback_called = False
        self.last_variable = None
        self.last_request = None

    def callback(self, request, variable):
        self.callback_called = True
        self.last_variable = variable
        self.last_request = request

    def test_route_callback(self):
        req = Request.blank(u"static/VALUE/morestatic")
        self.r(req, variable=u"VALUE")
        self.assertTrue(self.callback_called)
        self.assertTrue(self.last_request is req)
        self.assertEqual(u"VALUE", self.last_variable)

    def test_route_make_url(self):
        self.assertEqual(u"static/VAR/morestatic",
                self.r.make_url(variable=u"VAR"))

    def test_route_match_url(self):
        self.assertEqual(None, self.r.match_url(u"static/morestatic"))
        v = self.r.match_url(u"static/foo/morestatic")
        self.assertTrue(isinstance(v, dict))
        self.assertTrue("variable" in v)
        self.assertEqual(u"foo", v["variable"])

    def test_route_validate_valid_data(self):
        valid_data = (
            u"VAR",  # Unicode string.
            u"FOO",  # Normal string should be converted to Unicode.
            u"a_b",  # Underscores should work.
            u"_123", # Also, numbers preceded with underscores.
        )
        for expected in valid_data:
            url = u"static/{}/morestatic".format(expected)
            v = self.r.validate_url(url)
            self.assertTrue(isinstance(v, dict))
            self.assertTrue("variable" in v)
            self.assertEqual(expected, v["variable"])

    def test_route_validate_invalid_data(self):
        invalid_data = (
            4.12,    # Float numbers must not match.
            123,     # Same with numbers.
            "a.b",   # Dotted identifiers.
            "-+:\\", # Dashes and other symbols.
        )
        for data in invalid_data:
            url = u"static/{}/morestatic".format(data)
            v = self.r.validate_url(url)
            self.assertEqual(None, v)

    def test_route_multiple_methods(self):
        r = routing.Route(self.callback,
                u"get-or-post/{variable}",
                ("GET", "POST"), "testroute")
        self.assertEqual(["GET", "POST"], sorted(r.methods))


class TestRouteSchemaValidation(unittest.TestCase):

    test_data = {
        "int": {
            "valid": (12, -5),
            "invalid": (2.34, -3.4, 1e-34, "blah"),
        },
        "uint": {
            "valid": (0, 1, 23, 42, 1000101),
            "invalid": (-1, -43, -3.4, 2.34, 1e-34, "blah"),
        },
        "str": {
            "valid": (
                u"☺→",   # Some funky Unicode
                u"foo",  # Normal stuff
                u":-\\", # Arbitrary non-slash ASCII characters
            ),
            "invalid": (),
        },
        "dotid": {
            "valid": (
                u"foo.bar",
                u"foo.123",
                u"_24.foo",
                u"_23.321",
                u"_",
                u"_1",
            ),
            "invalid": (
                u"1",
                u"123.321",
            ),
        },
    }

    @staticmethod
    def make_test_valid(typename, value):
        fname = u"test_valid_{}_{!s}".format(typename, value)
        def f(self):
            r = routing.Route(lambda *arg, **kw: None,
                    u"item/{{value:{}}}".format(typename),
                    "GET", "testroute")
            v = r.validate_url(u"item/{!s}".format(value))
            self.assertTrue(isinstance(v, dict))
            self.assertTrue("value" in v)
            self.assertEqual(value, v["value"])
        return sanitize_ident(fname), f

    @staticmethod
    def make_test_invalid(typename, value):
        fname = u"test_invalid_{}_{!s}".format(typename, value)
        def f(self):
            r = routing.Route(lambda *arg, **kw: None,
                    u"item/{{value:{}}}".format(typename),
                    "GET", "testroute")
            # Either validation may return None, if the URL is
            # not matched, or the conversion from strings to
            # Python typed fails with valid.SchemaError.
            try:
                v = r.validate_url(u"item/{!s}".format(value))
                self.assertEqual(None, v)
            except valid.SchemaError:
                pass
        return sanitize_ident(fname), f

    @classmethod
    def inject_test_functions(cls):
        for name, kinds in iteritems(cls.test_data):
            for value in kinds["valid"]:
                setattr(cls, *cls.make_test_valid(name, value))
            for value in kinds["invalid"]:
                setattr(cls, *cls.make_test_invalid(name, value))

TestRouteSchemaValidation.inject_test_functions()



class TestableResource(routing.Routes, routing.Dispatcher):

    @routing.get("get/{variable}")
    def get_variable(self, request, variable):
        return self, request, variable

    @routing.route("head/{variable}", method="HEAD")
    def head_variable(self, request, variable):
        return self, request, variable

    @routing.post("post/{variable}")
    def post_variable(self, request, variable):
        return self, request, variable

    @routing.patch("patch/{variable}")
    def patch_variable(self, request, variable):
        return self, request, variable

    @routing.put("put/{variable}")
    def put_variable(self, request, variable):
        return self, request, variable

    @routing.delete("delete/{variable}")
    def delete_variable(self, request, variable):
        return self, request, variable

    @routing.route("any/{variable}")
    def any_variable(self, request, variable):
        return self, request, variable


class TestResource(unittest.TestCase):

    def setUp(self):
        self.rsrc = TestableResource()

    @staticmethod
    def make_method_dispatch_test(method):
        url = method.lower() + "/foo"
        def f(self):
            req = Request.blank(url, {"REQUEST_METHOD": method})
            r, rreq, rvar = self.rsrc.dispatch_request(req)
            self.assertEqual(self.rsrc, r)
            self.assertEqual(req, rreq)
            self.assertEqual("foo", rvar)
        return "test_dispatch_method_" + method, f

    @staticmethod
    def make_method_dispatch_any_test(method):
        def f(self):
            req = Request.blank("any/bar", {"REQUEST_METHOD": method})
            r, rreq, rvar = self.rsrc.dispatch_request(req)
            self.assertEqual(self.rsrc, r)
            self.assertEqual(req, rreq)
            self.assertEqual("bar", rvar)
        return "test_dispatch_method_any_" + method, f

    @classmethod
    def inject_test_functions(cls):
        for method in ("get", "post", "patch", "put", "delete", "head"):
            setattr(cls, *cls.make_method_dispatch_test(method))
            setattr(cls, *cls.make_method_dispatch_any_test(method))

    def test_dispatch_head_to_get(self):
        req = Request.blank("get/thisishead", {"REQUEST_METHOD": "HEAD"})
        r, rreq, rvar = self.rsrc.dispatch_request(req)
        self.assertEqual(self.rsrc, r)
        self.assertEqual(req, rreq)
        self.assertEqual("thisishead", rvar)

    def test_dispatch_not_found(self):
        req = Request.blank("/")
        with self.assertRaises(HTTPNotFound):
            self.rsrc.dispatch_request(req)

    def test_dispatch_invalid_method(self):
        req = Request.blank("get/bar", {"REQUEST_METHOD": "PATCH"})
        with self.assertRaises(HTTPMethodNotAllowed):
            self.rsrc.dispatch_request(req)

TestResource.inject_test_functions()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the MIT license.

"""
Authenticates a single user with a fixed password.

An username and password pair is kept in memory to be checked against. A
typical usage of this store is to provide a single user that is known only
by OMNI; for example, to define an user with administrative privileges that
can access the OMNI web interface without restrictions:

.. code-block:: lighttpd
   :linenos:

    stores {
        trivial.omni-admin {
            username "baron"
            password "redtriplane"
        }

        # Other stores...
    }

    http {
        web {
            admin "trivial.omni-admin"
        }
    }


**Configuration options:**

``username``
    Fixed user name.

``password`` (optional)
    Password for the user, in plain text.
"""

from .. import store, valid


class TrivialStore(store.Base):
    readonly = False

    def __init__(self, username, password):
        super(TrivialStore, self).__init__()
        self._credentials = (username, password)

    def authenticate(self, username, password):
        return (username, password) == self._credentials

    def usernames(self):
        yield self._credentials[0]

    def set_password(self, username, password):
        if username != self._credentials[0]:
            raise KeyError("invalid username {}".format(username))
        self._credentials = (username, password)


config_schema = valid.Schema({
    "username": valid.Identifier,
    valid.Optional("password"): valid.Password,
})


@valid.argument(config_schema)
def from_config(config):
    return TrivialStore(config["username"], config.get("password", ""))



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the GPLv3 license.

"""
Command line entry point for the OMNI daemon.
"""

from six import iteritems, string_types, text_type
from . import store, valid


config_realms_schema = valid.Schema({
    "methods": [valid.DotIdentifier],
    valid.Optional("description"): valid.Text,
})


def load_config(path):
    with open(path, "rb") as f:
        import wcfg
        return wcfg.load(f)


def make_application(omni_config):
    app = store.OMNI()
    for name, store_config in iteritems(omni_config.get("stores", {})):
        store_type = name.split(".", 1)[0]
        store_module = store.find(store_type)
        try:
            store_config = store_module.config_schema.validate(store_config)
        except valid.SchemaError as e:
            raise valid.SchemaError((), "in 'stores' section, item '{}': {!s}"
                    .format(name, e))
        app.add_store(name, store_module.from_config(store_config))

    for name, realm_config in iteritems(omni_config.get("realms", {})):
        try:
            realm_config = config_realms_schema.validate(realm_config)
        except valid.SchemaError as e:
            raise valid.SchemaError((), "in 'realms' section: {!s}"
                    .format(e))

        methods = realm_config["methods"]
        try:
            realm = store.Realm(realm_config.get("description", name),
                    (app.get_store(name) for name in methods))
        except KeyError as e:
            raise valid.SchemaError((), ("in 'realms' section, item '{s}':"
                " undefined store '{!s}'").format(name, e))
        app.add_realm(name, realm)

    return app


def make_wsgi_application(app_or_config):
    from omni.httpserver import WSGIApplication

    if not isinstance(app_or_config, store.OMNI):
        if isinstance(app_or_config, string_types + (text_type,)):
            app_or_config = load_config(app_or_config)
        app_or_config = make_application(app_or_config)

    assert isinstance(app_or_config, store.OMNI)
    return WSGIApplication(app_or_config)


class uWSGIApplication(object):
    def __call__(self, environ, start_response):
        import uwsgi
        config_file_path = uwsgi.opt["omni_config_file"]
        app = make_wsgi_application(load_config(config_file_path))
        # Replace the method itself, so make_wsgi_application()
        # is only called the first time, and the subsequent calls
        # go directly to the WSGI application itself.
        self.__call__ = app
        return app(environ, start_response)

uwsgi_application = uWSGIApplication()





#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the MIT license.

from setuptools import setup, find_packages
from setuptools import find_packages
from codecs import open
from os import path
import sys

# Prefer the local version of omni.metadata
sys.path.insert(0, path.abspath(path.dirname(__file__)))
from omni.metadata import metadata


def file_contents(*relpath):
    with open(path.join(path.dirname(__file__), *relpath), "rU",
            encoding="utf-8") as f:
        return f.read()


if __name__ == "__main__":
    setup(
        name=metadata.package,
        version=metadata.version,
        description=metadata.description,
        long_description=file_contents("README.rst"),
        author=metadata.main_author_name,
        author_email=metadata.main_author_email,
        url=metadata.url,
        packages=find_packages(),
        tests_require=metadata.test_requirements,
        install_requires=metadata.requirements,
        extras_require=metadata.extra_requirements,
        license=metadata.license,
        classifiers=metadata.classifiers,
        test_suite="omni.test",
        include_package_data=True,
        entry_points={ "console_scripts": metadata.entry_points },
    )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 Adrian Perez <aperez@igalia.com>
#
# Distributed under terms of the GPLv3 license.

"""
Provides the root resource of the OMNI Web User Interface

Route Template Syntax
---------------------

Route template strings provide a pattern with wildcards used to match URLs.
URL portions matching wildcards are extracted from the URL and converted
according to optional type annotations.

A template may not specify any wildcards. In that case the URL is matched
literally (e.g. ``foo/bar``).

Wildcards are names enclosed in brackets, e.g. ``{name}``. A type
annotation may optionally follow after a colon, e.g. ``{name:int}``.
Names must be unique for each route, and they must be valid Python
identifiers. Type annotations also determine how values are converted
before being passed to the route callback. The following type
annotations are available:

========== ================= ==============
Annotation Type              Converted Type
========== ================= ==============
``id``     Identifier        ``unicode``
           *(default)*
---------- ----------------- --------------
``dotid``  Identifier which  ``unicode``
           may contain dots
---------- ----------------- --------------
``str``    Arbitrary text    ``unicode``
---------- ----------------- --------------
``int``    Integer           ``int``
---------- ----------------- --------------
``uint``   Unsigned integer  ``int``
========== ================= ==============


As an example, the template ``{item}/{action}`` matches as follows:

========== =======================================
Path       Result
========== =======================================
/123/save  ``{"item": u"123", "action": u"save"}``
/123/save/ *No match.*
/123/      *No match.*
//save     *No match.*
========== =======================================


Module Contents
---------------

"""

from .. import valid
from functools import wraps
from webob.compat import text_type, bytes_
from webob.exc import HTTPNotFound, HTTPNotAcceptable, HTTPMethodNotAllowed
from webob.exc import HTTPUnauthorized, HTTPException
from webob import Request, Response
from base64 import b64decode
from six import itervalues
import re, inspect

camel_to_under_first_cap_sub = re.compile(r"(.)([A-Z][a-z]+)").sub
camel_to_under_all_cap_sub = re.compile(r"([a-z0-9])([A-Z])").sub
camel_to_under = lambda value: \
    camel_to_under_all_cap_sub(r"\1_\2", \
        camel_to_under_first_cap_sub(r"\1_\2", value)).lower()
template_var_re_finditer = \
    re.compile(r"\{(?P<name>[_a-zA-Z][_\w]*)(:(?P<type>[_\w^\d][_\w]*))?\}") \
        .finditer

_schema_map = {
    "int"   : ("d", "(?P<{}>[-+]?\d+)",          valid.Number),
    "uint"  : ("d", "(?P<{}>\d+)",               valid.NaturalNumber),
    "str"   : ("s", "(?P<{}>[^/]+)",             valid.Text),
    "id"    : ("s", "(?P<{}>[_a-zA-Z][_\w]*)",   valid.Identifier),
    "dotid" : ("s", "(?P<{}>[_a-zA-Z][_\.\w]*)", valid.DotIdentifier),
}


def parse_route_template(template):
    rbuilder = ["^"]
    fbuilder = []
    position = 0
    schema = {}

    for match in template_var_re_finditer(template):
        param_name = match.group("name")
        param_type = match.group("type") or "id"
        # TODO: Handle KeyError, maybe we want to use a custom error here.
        param_formatchar, param_re, param_schema = _schema_map[param_type]
        schema[param_name] = param_schema

        rbuilder.append(re.escape(template[position:match.start()]))
        rbuilder.append(param_re.format(param_name))

        fbuilder.append(template[position:match.start()])
        fbuilder.append("{")
        fbuilder.append(param_name)
        fbuilder.append(":")
        fbuilder.append(param_formatchar)
        fbuilder.append("}")

        position = match.end()

    rbuilder.append(re.escape(template[position:]))
    rbuilder.append("$")
    fbuilder.append(template[position:])

    return (valid.Schema(schema),
            re.compile(u"".join(rbuilder)),
            u"".join(fbuilder).format)


def make_tuple(v):
    if isinstance(v, (list, tuple)) or inspect.isgenerator(v):
        return tuple(v)
    else:
        return (v,)


class Route(object):
    """
    Defines a callback to be invoked for a certain URL template.

    See `route template syntax`_ for details on the mini-language accepted
    in URL template strings.

    :param callback: Callable to invoke when the route is matched.
    :param template: Route template string.
    :param methods: HTTP methods handled by this route.
    :param name: Name of the route. May be used to look the route back
        by name.
    """
    def __init__(self, callback, template, methods, name=None):
        s, r, f = parse_route_template(template)
        self.methods  = tuple(sorted((m.upper() for m in make_tuple(methods))))
        self.callback = callback
        self.template = template
        self.name     = name
        self._schema  = s
        self._format  = f
        self._match   = r.match

    def __eq__(self, other):
        # XXX: Probably we want to compare a version of the regexp stripped
        #      of the capture names, to determine whether both routes would
        #      match the same set of URLs.
        return self is other or (isinstance(other, Route)
                and self.methods == other.methods
                and self.template == other.template)

    def __repr__(self):  # pragma: no cover
        return "Route({!r}, {!r}, name={!r})".format(self.template,
                self.methods[0] if len(self.methods) == 1 else self.methods,
                self.name)

    def __call__(self, request, *arg, **kw):
        """
        Invokes the route callback.

        :param request: A :class:`webob.Request` instance.
        """
        return self.callback(request, *arg, **kw)

    def make_url(self, base="", **kw):
        """
        Builds a matching URL for the route.

        This method allows to construct valid URLs that would match the
        route template. The passed keyword arguments are used to fill-in
        the wildcards from the template.

        :param base: Base URL used as prefix.
        :param kw: Parameter values.
        :return: URL as a string.
        """
        return base + self._format(**kw)

    def match_url(self, url):
        """
        Try to match an URL against the route template.

        Matches the given `url` agains the route template. On a successful
        match, a dictionary mapping wildcards to the values they matched in
        the `url` is returned. If the `url` does not match the template,
        `None` is returned instead.

        :param url: The URL to be matched.
        :returns: `None` (if unmatched) or `dict` (if matched).
        """
        match = self._match(url)
        return None if match is None else match.groupdict()

    def validate_data(self, data):
        """
        Validate and convert data according to annotations in the template.

        Uses the type annotations from the route template to validate
        a dictionary in which keys are the names of the wildcards, and their
        values are strings. A new dictionary is returned, with the values
        converted accordingly. Typically, this method is used on the data
        dictionary returned by :func:`match_url()`.

        See `route template syntax`_ for details on how conversions are
        performed on the input data.

        If validation fails :class:`omni.valid.SchemaError` is raised.

        :param data: Dictionary with data.
        :returns: New dictionary with data validated and converted.
        """
        return self._schema.validate(data)

    def validate_url(self, url):
        """
        Matches, validates and convert data from an URL in a single step.

        This is equivalent to use :func:`match_url()` followed by
        :func:`validate_data()`. The returned dictionary will contain
        values which are already converted.

        :param url: The URL to be matched.
        :returns: `None` (if unmatched) or `dict` (if matched).
        """
        match = self._match(url)
        if match:
            return self.validate_data(match.groupdict())
        return None


def route(template, method="ANY", name=None):
    """
    Decorates a function as a route.

    :param template: Route template (see `route template syntax`_).
    :param method: HTTP methods for the route, either a single value
        or a list of strings.
    :param name: Name of the route. If not provided, the name of the
        decorated function is used.
    """
    def partial(func):
        return wraps(func)(Route(func, template, method,
            camel_to_under(func.__name__) if name is None else name))
    return partial

def get(template, *arg, **kw):
    """
    Decorates a function as a route, using ``GET`` as method.

    See :func:`route()` for details.
    """
    return route(template, "GET", *arg, **kw)

def post(template, *arg, **kw):
    """
    Decorates a function as a route, using ``POST`` as method.

    See :func:`route()` for details.
    """
    return route(template, "POST", *arg, **kw)

def put(template, *arg, **kw):
    """
    Decorates a function as a route, using ``PUT`` as method.

    See :func:`route()` for details.
    """
    return route(template, "PUT", *arg, **kw)

def delete(template, *arg, **kw):
    """
    Decorates a function as a route, using ``DELETE`` as method.

    See :func:`route()` for details.
    """
    return route(template, "DELETE", *arg, **kw)

def patch(template, *arg, **kw):
    """
    Decorates a function as a route, using ``PATCH`` as method.

    See :func:`route()` for details.
    """
    return route(template, "PATCH", *arg, **kw)


def authenticate(get_realm, realm_param="realm"):
    """
    Decorate a request handler to require HTTP basic authentication.

    Typical usage:

    .. code-block:: python

        class Authenticated(Dispatcher):
            def obtain_realm(self):
                return # ...

            @get("/realm/{r}/{page}")
            @authenticate(obtain_realm, "r")
            def handle_request(self, request, page):
                return get_page(page)

    :param get_realm: Function or callable used to obtain a :class:`Realm`
        instance. It will be passed the realm name as retrieved from the
        `realm_param` parameter of the original function. Pass ``None`` to
        disable inspecting the parameters of the decorated function.
    :param realm_param: Name of the parameter which contains the name of
        the realm to use. Note that the parameter will not be passed to the
        decorated function.
    """
    def partial(func):
        @wraps(func)
        def f(self, request, **kw):
            realm = kw[realm_param]
            del kw[realm_param]
            realm = get_realm(self, realm)
            username = None
            password = None
            if request.authorization:
                password = b64decode(request.authorization[1])
                username, password = password.decode("utf-8").split(":", 1)
            if username is None or not realm.authenticate(username, password):
                raise HTTPUnauthorized(headers=[
                    ("WWW-Authenticate",
                        "Basic realm=\"{}\"".format(realm.description)),
                ])
            return func(self, request, **kw)
        return f
    return partial


def after(after_func, *after_arg, **after_kw):
    """
    Decorator that calls a function with the result of another.

    Decorates a function and arranges for `after_func` to be called with
    the return value of the decorated function. Additional arguments can
    be passed to `after_func` by passing them to the decorator.

    For example, the following can be used to automatically convert the
    values returned by the ``generate_data()`` into JSON:

    >>> import json
    >>> @after(json.dumps)
    >>> def generate_data():
    ...     return { 'a': 42, 'b' }
    ...
    >>> generate_data()
    '{"a":42}'

    :param after_func: Function to be run after the decorated function.
        It can be a callable object.
    :param after_arg: Additional arguments to be passed to `after_func`.
    :param after_kw: Additional keyword arguments to be passed to
        `after_func`.
    """
    def partial(func):
        @wraps(func)
        def f(*arg, **kw):
            return after_func(func(*arg, **kw),
                    *after_arg, **after_kw)
        return f
    return partial


class Dispatcher(object):
    """
    Handles dispatching of HTTP requests.
    """
    def __init__(self):
        self.__route_dispatch = {}
        # Give subclasses a chance to override "routes" and get their
        # routes automagically plugged in at instance initialization.
        self.plug_routes(self.routes)

    @property
    def routes(self):
        """
        Enumerate all the `Route` objects known by the dispatcher.

        Accessing this property returns a generator which yields all the
        routes known by this dispatcher.
        """
        seen = set()
        for routes in itervalues(self.__route_dispatch):
            for r in routes:
                if r in seen:
                    continue
                seen.add(r)
                yield r

    def add_route(self, route):
        """
        Adds a route to the dispatcher.

        :param route: An instance of :class:`Route`
        """
        assert isinstance(route, Route)
        # TODO: Check that existing routes won't shadow the URLs
        #       matched by the one being added.
        for method in route.methods:
            if method not in self.__route_dispatch:
                self.__route_dispatch[method] = []
            self.__route_dispatch[method].append(route)

    def plug_routes(self, routes):
        """
        Imports all the routes from another dispatcher.

        Picks the routes known by `other` and plugs them into the dispatcher

        :param routes: Iterable containing :class:`Route` instances.
        """
        for r in routes:
            self.add_route(r)

    def url(self, route=None, base=None, **kw):
        """
        Builds an URL for a given route.

        Note that `route` is not supplied, the name of the caller will
        be used as the route name. This is provided as a shorthand for
        those routes which want to refer to themselves.

        If `base` is not supplied, if there is a local variable in the
        frame of the caller named ``request`` which is an instance of
        :class:`webob.Request`, it will be used as if the request was
        passed as `base`.

        :param route: Name of the route, a :class:`Route` instance, or
            ``None``.
        :param base: Base URL, either a string, a :class:`webob.Request`
            instance from which to infer the base URL, or ``None``.
        :param kw: Keyword arguments to fill-in the wildcards in the
            route URL template.
        :return: A string.
        """
        # TODO: This is potentially slow, check whether there should be
        #       a mapping from names to routes, or some kind of caching.
        if base is None and isinstance(route, Request):
            base = route
            route = None

        if route is None or base is None:
            from sys import _getframe
            frame = _getframe(1)
            if route is None:
                route = frame.f_code.co_name
            if base is None:
                base = frame.f_locals.get("request", None)
                if not isinstance(base, Request):
                    base = None

        if isinstance(base, Request):
            base = base.script_name
        if isinstance(route, Route):
            return route.make_url(base, **kw)

        for r in self.routes:
            if r.name == route:
                return r.make_url(base, **kw)
        raise KeyError(route)

    def dispatch_request(self, request):
        """
        Dispatches a request to the known routes.

        :param request: A :class:`webob.Request` instance.
        """
        method = request.method.upper()
        path = request.path_info or "/"
        if method == "HEAD":
            try_methods = ("PROXY", method, "GET", "ANY")
        else:
            try_methods = ("PROXY", method, "ANY")

        tried_methods = set()
        for m in try_methods:
            if m in self.__route_dispatch:
                tried_methods.add(m)
                for r in self.__route_dispatch[m]:
                    try:
                        data = r.validate_url(path)
                    except valid.SchemaError as e:
                        raise HTTPNotAcceptable(e)
                    if data:
                        return r(self, request, **data)

        allowed = set()
        tried = set(try_methods)
        for method in set(self.__route_dispatch) - tried:
            for r in self.__route_dispatch[method]:
                if r.match_url(path):
                    allowed.add(method)

        if allowed:
            raise HTTPMethodNotAllowed(headers=[
                ("Allow", ",".join(allowed)),
            ])
        raise HTTPNotFound()

    def dispatch_wsgi(self, environ, start_response):
        """
        Dispatches a WSGI request.

        :param environ: WSGI request environment.
        :param start_response: WSGI response callback
        :returns: A valid WSGI response content.
        """
        req = Request(environ)
        req.response = Response()

        try:
            resp = self.dispatch_request(req)
        except HTTPException as e:
            resp = e

        if inspect.isgenerator(resp):
            resp = Response(app_iter=resp)
        elif resp is None:
            resp = req.response
        if isinstance(resp, text_type):
            resp = bytes_(resp, req.charset)
        if isinstance(resp, bytes):
            body = resp
            resp = req.response
            resp.write(body)
        if resp is not req.response:
            resp = req.response.merge_cookies(resp)
        return resp(environ, start_response)


class Routes(object):
    @property
    def routes(self):
        for _, r in inspect.getmembers(self,
                lambda r: isinstance(r, Route)):
            yield r


__all__ = [
    "Route", "Dispatcher", "Routes",
    "route", "get", "post", "put", "delete", "patch",
    "after", "authenticate",
]
